---------------------------------------------------------------------
--  PhaseX - Copyright 2015 Ranfan@Blackrock-EU (candy@taomi.net)  --
-- All rights reserved. Unauthorized redistribution is prohibited. --
---------------------------------------------------------------------

-- Changelog:
-- 0.9: Adjusted phase switch for possible boss healing, tested more bosses, fixed some bosses
-- 0.8: Tested on first 8 HFC bosses (up til Shadow-Lord Iskar, Soulbound Construct), fixed errors
-- 0.7: Added calculations and mechanics for HFC boss encounters based on boss guides
-- 0.6: Broke down code to multiple files and refactored functions to ease fixing and development, allowing different calculations for different fights
-- 0.5: Modified visibility to only show the bar during boss encounters which make use of the boss raid frames
-- 0.4: Modified to multi-layer bar to support multiple phases during a fight
-- 0.3: Adjusted bar visibility based on player status (combat/non-combat)
-- 0.2: Added timer to show the estimated end of the phase
-- 0.1: Phase bar based on health

-- Todo:
-- * Test addon on remaining boss encounters for errors and glitches
-- * Implement console commands to move/resize the bar and change settings
-- * Implement configuration save/reset function
-- * [low priority] Add support for older raid tiers (currently only HFC supported)

local config = PhaseX_Config
local frame = PhaseX_Frame

frame.SetupBoss = function()
	local bossList = frame.GetBossList()
	if (bossList[90019] ~= nil) then -- Hellfire Assault
		return frame.Setup_WOD_HFC_01()
	end
	if (bossList[90284] ~= nil) then -- Iron Reaver
		return frame.Setup_WOD_HFC_02()
	end
	if (bossList[90435] ~= nil) then -- Kormrok
		return frame.Setup_WOD_HFC_03()
	end
	if (bossList[92142] ~= nil or bossList[92144] ~= nil or bossList[92146] ~= nil) then -- Hellfire High Council
		return frame.Setup_WOD_HFC_04()
	end
	if (bossList[90378] ~= nil) then -- Kilrogg Deadeye
		return frame.Setup_WOD_HFC_05()
	end
	if (bossList[90199] ~= nil) then -- Gorefiend
		return frame.Setup_WOD_HFC_06()
	end
	if (bossList[90316] ~= nil) then -- Shadow-Lord Iskar
		return frame.Setup_WOD_HFC_07()
	end
	if (bossList[90296] ~= nil or bossList[92330] ~= nil) then -- Soulbound Construct
		return frame.Setup_WOD_HFC_08()
	end
	if (bossList[90269] ~= nil) then -- Tyrant Velhari
		return frame.Setup_WOD_HFC_09()
	end
	if (bossList[89890] ~= nil) then -- Fel Lord Zakuun
		return frame.Setup_WOD_HFC_10()
	end
	if (bossList[93068] ~= nil) then -- Xhul'horac
		return frame.Setup_WOD_HFC_11()
	end
	if (bossList[91305] ~= nil or bossList[91349] ~= nil) then -- Mannoroth
		return frame.Setup_WOD_HFC_12()
	end
	if (bossList[91331] ~= nil) then -- Archimonde
		return frame.Setup_WOD_HFC_13()
	end
	if (#bossList > 0) then -- Unknown Boss
		frame.Setup_Default()
		frame.EncounterId = 1
		frame.EncounterUnit = "boss1"
	end
end

frame.Setup_WOD_HFC_01 = function() -- Hellfire Assault
	frame.Setup_Default()
	frame.EncounterId = 90019
	frame.EncounterUnit = "boss1"
end

frame.Setup_WOD_HFC_02 = function() -- Iron Reaver
	frame.EncounterId = 90284
	frame.EncounterUnit = "boss1"
	frame.PercentValue = 1
	frame.PhaseCurrent = 1
	frame.PhaseMax = 2
	frame.PhaseStart = GetTime()
	frame.UpdateBoss = function() end
	frame.UpdateEvent = function(event, ...)
		local arg = {...}
		local unitId = arg[1]
		local spellId = arg[5]
		if (event == "UNIT_SPELLCAST_SUCCEEDED" and unitId == "boss1" and spellId == 182055) then -- Full Charge
			frame.UpdatePhase(2)
		end
		if (event == "UNIT_SPELLCAST_SUCCEEDED" and unitId == "boss1" and spellId == 182066) then -- Falling Slam
			frame.UpdatePhase(1)
		end
	end
	frame.UpdateStatus = function()
		if (frame.PhaseCurrent == 1) then -- Phase 1, Energy filling up
			frame.PercentValue = 1 - frame.GetPercentValue(UnitPower(frame.EncounterUnit), UnitPowerMax(frame.EncounterUnit))
		end
		if (frame.PhaseCurrent == 2) then -- Phase 2, Timer 60s
			frame.PercentValue = 1 - frame.GetPercentValue(GetTime() - frame.PhaseStart, 60)
		end
	end
end

frame.Setup_WOD_HFC_03 = function() -- Kormrok
	frame.Setup_Default()
	frame.EncounterId = 90435
	frame.EncounterUnit = "boss1"
end

frame.Setup_WOD_HFC_04 = function() -- Hellfire High Council
	frame.EncounterId = 90435
	frame.EncounterUnit = "boss1"
	frame.PercentValue = 1
	frame.PhaseCurrent = 1
	frame.PhaseMax = 2
	frame.PhaseStart = GetTime()
	frame.Temp_HellfireHighCouncil = 0
	frame.UpdateBoss = function() end
	frame.UpdateEvent = function(event, ...) end
	frame.UpdateStatus = function()
		local boss1 = frame.GetPercentValue(UnitHealth("boss1"), UnitHealthMax("boss1"))
		local boss2 = frame.GetPercentValue(UnitHealth("boss2"), UnitHealthMax("boss2"))
		if (UnitExists("boss2")) then -- Cache health of vanishing boss
			frame.Temp_HellfireHighCouncil = boss2
		else
			boss2 = frame.Temp_HellfireHighCouncil
		end
		local boss3 = frame.GetPercentValue(UnitHealth("boss3"), UnitHealthMax("boss3"))
		local list_normal = {1}
		local list_enrage = {0}
		if (boss1 >= 0.3) then
			list_normal[#list_normal+1] = boss1
		elseif (boss1 > 0) then
			list_enrage[#list_enrage+1] = boss1
		end
		if (boss2 >= 0.3) then
			list_normal[#list_normal+1] = boss2
		elseif (boss2 > 0.01) then
			list_enrage[#list_enrage+1] = boss2
		end
		if (boss3 >= 0.3) then
			list_normal[#list_normal+1] = boss3
		elseif (boss3 > 0) then
			list_enrage[#list_enrage+1] = boss3
		end
		if (#list_enrage > 1) then -- At least one boss enraged
			frame.UpdatePhase(2)
			frame.PercentValue = max(unpack(list_enrage))/0.3
		else -- No boss enraged
			frame.UpdatePhase(1)
			frame.PercentValue = (min(unpack(list_normal))-0.3)/0.7
		end
	end
end

frame.Setup_WOD_HFC_05 = function() -- Kilrogg Deadeye
	frame.Setup_Default()
	frame.EncounterId = 90378
	frame.EncounterUnit = "boss1"
end

frame.Setup_WOD_HFC_06 = function() -- Gorefiend
	frame.EncounterId = 90199
	frame.EncounterUnit = "boss1"
	frame.PercentValue = 1
	frame.PhaseCurrent = 1
	frame.PhaseMax = 2
	frame.PhaseStart = GetTime()
	frame.UpdateBoss = function() end
	frame.UpdateEvent = function(event, ...)
		local arg = {...}
		local unitId = arg[1]
		local spellId = arg[5]
		if (event == "UNIT_SPELLCAST_SUCCEEDED" and unitId == "boss1" and spellId == 181973) then -- Feast of Souls
			frame.UpdatePhase(2)
			frame.PercentValue = 1
		end
	end
	frame.UpdateStatus = function()
		if (frame.PhaseCurrent == 2) then -- Phase 2, Energy filling up
			frame.PercentValue = 1 - frame.GetPercentValue(UnitPower(frame.EncounterUnit), UnitPowerMax(frame.EncounterUnit))
			if (frame.PercentValue == 0) then
				frame.UpdatePhase(1)
				frame.PercentValue = 1
			end
		end
		if (frame.PhaseCurrent == 1) then -- Phase 1, Energy going down
			frame.PercentValue = frame.GetPercentValue(UnitPower(frame.EncounterUnit), UnitPowerMax(frame.EncounterUnit))
		end
	end
end

frame.Setup_WOD_HFC_07 = function() -- Shadow-Lord Iskar
	frame.EncounterId = 90316
	frame.EncounterUnit = "boss1"
	frame.PercentValue = 1
	frame.PhaseCurrent = 1
	frame.PhaseMax = 7
	frame.PhaseStart = GetTime()
	frame.UpdateBoss = function() end
	frame.UpdateEvent = function(event, ...)
		local arg = {...}
		local unitId = arg[1]
		local spellId = arg[5]
		if (event == "UNIT_SPELLCAST_SUCCEEDED" and unitId == "boss1" and spellId == 181873) then -- Shadow Escape
			frame.UpdatePhase(frame.PhaseCurrent+1)
		end
	end
	frame.UpdateStatus = function()
		if (frame.PhaseCurrent == 1) then -- Phase 1, Health from 100% to 70%
			frame.PercentValue = (frame.GetPercentValue(UnitHealth(frame.EncounterUnit), UnitHealthMax(frame.EncounterUnit))-0.7)/0.3
		end
		if (frame.PhaseCurrent == 3) then -- Phase 3, Health from 70% to 45%
			frame.PercentValue = (frame.GetPercentValue(UnitHealth(frame.EncounterUnit), UnitHealthMax(frame.EncounterUnit))-0.45)/0.25
		end
		if (frame.PhaseCurrent == 5) then -- Phase 5, Health from 45% to 20%
			frame.PercentValue = (frame.GetPercentValue(UnitHealth(frame.EncounterUnit), UnitHealthMax(frame.EncounterUnit))-0.20)/0.25
		end
		if (frame.PhaseCurrent == 7) then -- Phase 7, Health from 20% to 0%
			frame.PercentValue = (frame.GetPercentValue(UnitHealth(frame.EncounterUnit), UnitHealthMax(frame.EncounterUnit)))/0.2
		end
		if (frame.PhaseCurrent == 2 or frame.PhaseCurrent == 4 or frame.PhaseCurrent == 6) then -- Phase 2/4/6, Timer 40s
			frame.PercentValue = 1 - frame.GetPercentValue(GetTime() - frame.PhaseStart, 40)
			if (frame.PercentValue == 0) then
				frame.UpdatePhase(frame.PhaseCurrent+1)
				frame.PercentValue = 1
			end
		end
	end
end

frame.Setup_WOD_HFC_08 = function() -- Soulbound Construct
	frame.EncounterId = 90296
	frame.EncounterUnit = "boss1"
	frame.PercentValue = 1
	frame.PhaseCurrent = 1
	frame.PhaseMax = 2
	frame.PhaseStart = GetTime()
	frame.UpdateBoss = function() end
	frame.UpdateEvent = function(event, ...)
		local arg = {...}
		local unitId = arg[1]
		local spellId = arg[5]
		if (event == "UNIT_SPELLCAST_SUCCEEDED" and spellId == 183023 and (unitId == "boss1" or unitId == "boss2")) then -- Eject Soul
			if (frame.PhaseCurrent == 1) then
				frame.UpdatePhase(2)
				frame.EncounterUnit = "boss2"
			else
				frame.UpdatePhase(1)
				frame.EncounterUnit = "boss1"
			end
		end
	end
	frame.UpdateStatus = function()
		frame.PercentValue = frame.GetPercentValue(UnitHealth(frame.EncounterUnit), UnitHealthMax(frame.EncounterUnit))
	end
end

frame.Setup_WOD_HFC_09 = function() -- Tyrant Velhari
	frame.EncounterId = 90269
	frame.EncounterUnit = "boss1"
	frame.PercentValue = 1
	frame.PhaseCurrent = 1
	frame.PhaseMax = 3
	frame.PhaseStart = GetTime()
	frame.UpdateBoss = function() end
	frame.UpdateEvent = function() end
	frame.UpdateStatus = function()
		local boss = frame.GetPercentValue(UnitHealth(frame.EncounterUnit), UnitHealthMax(frame.EncounterUnit))
		if (boss <= 0.7 and frame.PhaseCurrent < 2) then
			frame.UpdatePhase(2)
		end
		if (boss <= 0.4 and frame.PhaseCurrent < 3) then
			frame.UpdatePhase(3)
		end
		if (frame.PhaseCurrent == 1) then -- Phase 1, 100% to 70%
			frame.PercentValue = (boss-0.7)/0.3
		elseif (frame.PhaseCurrent == 2) then -- Phase 2, 70% to 40%
			frame.PercentValue = (boss-0.4)/0.3
		else -- Phase 3, 40% to 0%
			frame.PercentValue = boss/0.4
		end
	end
end

frame.Setup_WOD_HFC_10 = function() -- Fel Lord Zakuun
	frame.EncounterId = 89890
	frame.EncounterUnit = "boss1"
	frame.PercentValue = 1
	frame.PhaseCurrent = 1
	frame.PhaseMax = 3
	frame.PhaseStart = GetTime()
	frame.UpdateBoss = function() end
	frame.UpdateEvent = function(event, ...)
		local arg = {...}
		local unitId = arg[1]
		local spellId = arg[5]
		if (event == "UNIT_SPELLCAST_SUCCEEDED" and spellId == 179668 and unitId == "boss1") then -- Throw Axe
			frame.UpdatePhase(2)
		end
	end
	frame.UpdateStatus = function()
		local boss = frame.GetPercentValue(UnitHealth(frame.EncounterUnit), UnitHealthMax(frame.EncounterUnit))
		if (boss <= 0.3 and frame.PhaseCurrent < 3) then
			frame.UpdatePhase(3)
		end
		if (frame.PhaseCurrent < 3) then -- Phase 1 or 2
			if (frame.PhaseCurrent == 2) then
				frame.PercentValue = 1 - frame.GetPercentValue(GetTime() - frame.PhaseStart, 30)
				if (frame.PercentValue == 0) then
					frame.UpdatePhase(1)
				end
			end
			if (frame.PhaseCurrent == 1) then
				frame.PercentValue = 1 - frame.GetPercentValue(GetTime() - frame.PhaseStart, 90)
			end
		else -- Phase 3, 30% to 0%
			frame.PercentValue = boss/0.3
		end
	end
end

frame.Setup_WOD_HFC_11 = function() -- Xhul'horac
	frame.EncounterId = 93068
	frame.EncounterUnit = "boss1"
	frame.PercentValue = 1
	frame.PhaseCurrent = 1
	frame.PhaseMax = 4
	frame.PhaseStart = GetTime()
	frame.UpdateBoss = function() end
	frame.UpdateEvent = function() end
	frame.UpdateStatus = function()
		local boss = frame.GetPercentValue(UnitHealth(frame.EncounterUnit), UnitHealthMax(frame.EncounterUnit))
		if (boss <= 0.65 and frame.PhaseCurrent < 2) then
			frame.UpdatePhase(2)
		end
		if (boss <= 0.35 and frame.PhaseCurrent < 3) then
			frame.UpdatePhase(3)
		end
		if (boss <= 0.2 and frame.PhaseCurrent < 4) then
			frame.UpdatePhase(4)
		end
		if (frame.PhaseCurrent == 1) then -- Phase 1, 100% to 65%
			frame.PercentValue = (boss-0.65)/0.35
		elseif (frame.PhaseCurrent == 2) then -- Phase 2, 65% to 35%
			frame.PercentValue = (boss-0.35)/0.3
		elseif (frame.PhaseCurrent == 3) then -- Phase 3, 35% to 20%
			frame.PercentValue = (boss-0.2)/0.15
		else -- Phase 4, 20% to 0%
			frame.PercentValue = boss/0.2
		end
	end
end

frame.Setup_WOD_HFC_12 = function() -- Mannoroth
	frame.Setup_Default()
	frame.EncounterId = 91349
	frame.EncounterUnit = "boss1"
	frame.PercentValue = 1
	frame.PhaseCurrent = 1
	frame.PhaseMax = 4
	frame.PhaseStart = GetTime()
	frame.UpdateBoss = function() end
	frame.UpdateEvent = function() end
	frame.UpdateStatus = function()
		local bossList = frame.GetBossList()
		if (bossList[91349] == nil) then -- Phase 1
			local boss1 = frame.GetPercentValue(UnitHealth("boss1"), UnitHealthMax("boss1"))
			local boss2 = frame.GetPercentValue(UnitHealth("boss2"), UnitHealthMax("boss2"))
			local boss3 = frame.GetPercentValue(UnitHealth("boss3"), UnitHealthMax("boss3"))
			frame.PercentValue = (boss1+boss2+boss3)/3
		else -- Phase 2 to 4
			if (frame.PhaseCurrent < 2) then
				frame.UpdatePhase(2)
			end
			local boss = frame.GetPercentValue(UnitHealth(frame.EncounterUnit), UnitHealthMax(frame.EncounterUnit))
			if (boss <= 0.65 and frame.PhaseCurrent < 3) then
				frame.UpdatePhase(3)
			end
			if (boss <= 0.35 and frame.PhaseCurrent < 4) then
				frame.UpdatePhase(4)
			end
			if (frame.PhaseCurrent == 2) then -- Phase 2, 100% to 65%
				frame.PercentValue = (boss-0.65)/0.35
			elseif (frame.PhaseCurrent == 3) then -- Phase 3, 65% to 35%
				frame.PercentValue = (boss-0.35)/0.3
			else -- Phase 4, 35% to 0%
				frame.PercentValue = boss/0.35
			end
		end
	end
end

frame.Setup_WOD_HFC_13 = function() -- Archimonde
	frame.EncounterId = 91331
	frame.EncounterUnit = "boss1"
	frame.PercentValue = 1
	frame.PhaseCurrent = 1
	frame.PhaseMax = 6
	frame.PhaseStart = GetTime()
	frame.UpdateBoss = function() end
	frame.UpdateEvent = function() end
	frame.UpdateStatus = function()
		local boss = frame.GetPercentValue(UnitHealth(frame.EncounterUnit), UnitHealthMax(frame.EncounterUnit))
		if (boss <= 0.85 and frame.PhaseCurrent < 2) then
			frame.UpdatePhase(2)
		end
		if (boss <= 0.7 and frame.PhaseCurrent < 3) then
			frame.UpdatePhase(3)
		end
		if (boss <= 0.55 and frame.PhaseCurrent < 4) then
			frame.UpdatePhase(4)
		end
		if (boss <= 0.4 and frame.PhaseCurrent < 5) then
			frame.UpdatePhase(5)
		end
		if (boss <= 0.25 and frame.PhaseCurrent < 6) then
			frame.UpdatePhase(6)
		end
		if (frame.PhaseCurrent == 1) then -- Phase 1, 100% to 85%
			frame.PercentValue = (boss-0.85)/0.15
		elseif (frame.PhaseCurrent == 2) then -- Phase 2, 85% to 70%
			frame.PercentValue = (boss-0.7)/0.15
		elseif (frame.PhaseCurrent == 3) then -- Phase 3, 70% to 55%
			frame.PercentValue = (boss-0.55)/0.15
		elseif (frame.PhaseCurrent == 4) then -- Phase 4, 55% to 40%
			frame.PercentValue = (boss-0.4)/0.15
		elseif (frame.PhaseCurrent == 5) then -- Phase 5, 40% to 25%
			frame.PercentValue = (boss-0.25)/0.15
		else -- Phase 6, 25% to 0%
			frame.PercentValue = boss/0.25
		end
	end
end
